#![allow(unused)]

use std::fmt::Display;

mod exec;
mod inference;
mod parse;

type Ident = String;

#[derive(Debug, Clone, PartialEq, Eq)]
enum PrimitiveType {
    Nat,
    Bool,
}

#[derive(Debug, Clone, PartialEq, Eq)]
enum Type {
    Primitive(PrimitiveType),    // Bool | Nat
    Arrow(Box<Type>, Box<Type>), // 0 -> 1
}

impl From<PrimitiveType> for Type {
    fn from(value: PrimitiveType) -> Self {
        Type::Primitive(value)
    }
}

impl Type {
    fn arrow<T1: Into<Type>, T2: Into<Type>>(t1: T1, t2: T2) -> Self {
        Self::Arrow(Box::new(t1.into()), Box::new(t2.into()))
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
enum Constant {
    Nat(usize),
    Bool(bool),
}

#[derive(Debug, Clone, PartialEq, Eq)]
enum Ast {
    Abstraction(Ident, Type, Box<Ast>), // \0:1.2
    Application(Box<Ast>, Box<Ast>),    // 0 1
    Variable(Ident),                    // x
    Constant(Constant),                 // true | false | n
}

impl Display for Ast {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Ast::Abstraction(var, typ, ast) => write!(f, "(\\ ({var} {typ}) ({ast}))"),
            Ast::Application(lhs, rhs) => write!(f, "(lhs rhs)"),
            Ast::Variable(v) => write!(f, "{v}"),
            Ast::Constant(constant) => write!(f, "{constant}"),
        }
    }
}

impl Display for Type {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Type::Primitive(primitive_type) => write!(f, "{primitive_type}"),
            Type::Arrow(t1, t2) => write!(f, "({t1} -> {t2})"),
        }
    }
}

impl Display for PrimitiveType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            PrimitiveType::Nat => write!(f, "Nat"),
            PrimitiveType::Bool => write!(f, "Bool"),
        }
    }
}

impl Display for Constant {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Constant::Nat(n) => write!(f, "{n}"),
            Constant::Bool(b) => write!(f, "{b}"),
        }
    }
}
